/*
 * Java 4K 2011
 * Copyright (C) 2011 meatfighter.com
 *
 * This file is part of Java 4K 2011.
 *
 * Java 4K 2011 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java 4K 2011 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package main;

import java.applet.*;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.util.*;

public class Main extends Applet implements Runnable {

  private class Flash {
    public int tileX;
    public int tileY;
    public int delay = FLASH_DELAY;
    public int index;

    public Flash(int tileX, int tileY) {
      this.tileX = tileX;
      this.tileY = tileY;
    }
  }

  public static final int APPLET_WIDTH = 784;
  public static final int APPLET_HEIGHT = 512;
  public static final int IMAGE_WIDTH = 392;
  public static final int IMAGE_HEIGHT = 256;
  
  public static final int SPRITE_COUNT = 64;

  public static final int SPRITE_PADDLE = 50;
  public static final int SPRITE_BALL = 51;
  public static final int SPRITE_TEXT = 52;

  public static final int FLASH_DELAY = 5;

  public static final int[] FLASH_TILES = { 41, 42, 43, 45, 46 };

  private final BufferedImage[] sprites = new BufferedImage[SPRITE_COUNT];
  private final int[][] tiles = new int[32][49];
  private final ArrayList<Flash> flashes = new ArrayList<Flash>();

  private int ballX = 100;
  private int ballY = 100;
  private int ballVx = 1;
  private int ballVy = -1;
  private int paddleX = 0;
  private float angle = 0;

  @Override
  public void start() {
    new Thread(this).start();
  }

  public void run() {

    loadSprites();
    initializeTileMap();

    BufferedImage image = new BufferedImage(IMAGE_WIDTH, IMAGE_HEIGHT, 
        BufferedImage.TYPE_INT_RGB);
    Graphics2D g = (Graphics2D)image.getGraphics();
    Graphics2D g2 = null;

    long nextFrameStartTime = System.nanoTime();
    while(true) {

      do {
        nextFrameStartTime += 16666667;

        update();

      } while(nextFrameStartTime < System.nanoTime());

      render(g);

      // show the hidden buffer
      if (g2 == null) {
        g2 = (Graphics2D)getGraphics();
        requestFocus();
      } else {
        g2.drawImage(image, 0, 0, APPLET_WIDTH, APPLET_HEIGHT, null);
      }

      // burn off extra cycles
      while(nextFrameStartTime - System.nanoTime() > 0) {
        Thread.yield();
      }
    }
  }

  private void update() {

    // move ball
    for(int i = 0; i < 3; i++) {
      if (ballVx == 1) {
        int x = (ballX + 5) >> 3;
        int y = ballY >> 3;
        if (tiles[y][x] >= 32) {
          ballVx = -1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }

        y = (ballY + 3) >> 3;
        if (tiles[y][x] >= 32) {
          ballVx = -1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }
      } else {
        int x = (ballX - 1) >> 3;
        int y = ballY >> 3;
        if (tiles[y][x] >= 32) {
          ballVx = 1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }

        y = (ballY + 3) >> 3;
        if (tiles[y][x] >= 32) {
          ballVx = 1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }
      }
      ballX += ballVx;

      if (ballVy == 1) {

        int x = ballX >> 3;
        int y = (ballY + 4) >> 3;
        if (tiles[y][x] >= 32) {
          ballVy = -1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }

        x = (ballX + 4) >> 3;
        if (tiles[y][x] >= 32) {
          ballVy = -1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }

        if (ballY >= 228) {
          ballVy = -1;
        }
      } else {

        int x = ballX >> 3;
        int y = (ballY - 1) >> 3;
        if (tiles[y][x] >= 32) {
          ballVy = 1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }

        x = (ballX + 4) >> 3;
        if (tiles[y][x] >= 32) {
          ballVy = 1;
          if (tiles[y][x] == 40) {
            tiles[y][x] = 41;
            flashes.add(new Flash(x, y));
          }
        }
      }
      ballY += ballVy;
    }

    // update flashes
    for(int i = flashes.size() - 1; i >= 0; i--) {
      Flash flash = flashes.get(i);
      if (--flash.delay == 0) {
        flash.delay = FLASH_DELAY;
        if (++flash.index == 5) {
          tiles[flash.tileY][flash.tileX] = 40;
          flashes.remove(i);
          continue;
        } else {
          tiles[flash.tileY][flash.tileX] = FLASH_TILES[flash.index];
        }
      }
    }

    // move paddle
    angle += 0.1f;
    paddleX = ballX - 13 + (((int)((228 - ballY) * Math.sin(angle))) >> 2);
    if (paddleX < 8) {
      paddleX = 8;
    } else if (paddleX > 352) {
      paddleX = 352;
    }
  }

  private void render(Graphics2D g) {

    // draw tiles
    for(int y = 0; y < 32; y++) {
      for(int x = 0; x < 49; x++) {
        g.drawImage(sprites[tiles[y][x]], x << 3, y << 3, null);
      }
    }

    // draw ball
    g.drawImage(sprites[SPRITE_BALL], ballX, ballY, null);

    // draw paddle
    g.drawImage(sprites[SPRITE_PADDLE], paddleX, 232, null);
  }

  private void initializeTileMap() {
    for(int y = 0; y < 32; y++) {
      for(int x = 0; x < 49; x++) {
        tiles[y][x] = 4 * (y & 3) + (x & 3);        
      }
    }
    tiles[0][0] = 32;
    tiles[0][48] = 39;
    for(int x = 0; x < 47; x++) {
      tiles[0][x + 1] = 33;
    }
    for(int y = 0; y < 30; y += 5) {
      tiles[y + 1][48] = tiles[y + 1][0] = 38;
      tiles[y + 2][48] = tiles[y + 2][0] = 44;
      tiles[y + 3][48] = tiles[y + 3][0] = 47;
      tiles[y + 4][48] = tiles[y + 4][0] = 48;
      tiles[y + 5][48] = tiles[y + 5][0] = 49;
    }
    tiles[31][48] = tiles[31][0] = 38;
    tiles[0][10] = 34;
    tiles[0][11] = 35;
    tiles[0][12] = 36;
    tiles[0][13] = 37;

    tiles[0][35] = 34;
    tiles[0][36] = 35;
    tiles[0][37] = 36;
    tiles[0][38] = 37;
    for(int x = 0; x < 47; x++) {
      tiles[1][x + 1] += 16;
    }
    for(int y = 0; y < 30; y++) {
      tiles[y + 2][1] += 16;
    }

    for(int y = 0; y < 8; y++) {
      for(int x = 0; x < 43; x++) {
        if (sprites[SPRITE_TEXT].getRGB(x, y) == 0xFF000000) {
          tiles[y + 3][x + 3] = 40;
          if (tiles[y + 4][x + 4] < 16) {
            tiles[y + 4][x + 4] += 16;
          }
        }
      }
    }
  }

  private void loadSprites() {
    try {
      BufferedImage image
          = ImageIO.read(Main.class.getResource("/images/0.png"));
      for(int y = 0; y < 4; y++) {
        for(int x = 0; x < 4; x++) {
          sprites[y * 4 + x] = image.getSubimage(x << 3, y << 3, 8, 8);
        }
      }

      image = ImageIO.read(Main.class.getResource("/images/1.png"));
      for(int y = 0; y < 4; y++) {
        for(int x = 0; x < 4; x++) {
          sprites[16 + y * 4 + x] = image.getSubimage(x << 3, y << 3, 8, 8);
        }
      }

      image = ImageIO.read(Main.class.getResource("/images/2.png"));
      for(int y = 0; y < 3; y++) {
        for(int x = 0; x < 6; x++) {
          sprites[32 + y * 6 + x] = image.getSubimage(x << 3, y << 3, 8, 8);
        }
      }

      sprites[SPRITE_PADDLE]
          = ImageIO.read(Main.class.getResource("/images/3.png"));
      sprites[SPRITE_BALL]
          = ImageIO.read(Main.class.getResource("/images/4.png"));
      sprites[SPRITE_TEXT]
          = ImageIO.read(Main.class.getResource("/images/5.png"));

    } catch(Throwable t) {
      t.printStackTrace();
    }
  }

  public static void main(String[] args) throws Throwable {
    javax.swing.JFrame frame = new javax.swing.JFrame("Java 4K 2011");
    frame.setDefaultCloseOperation(javax.swing.JFrame.EXIT_ON_CLOSE);
    Main applet = new Main();
    applet.setPreferredSize(
        new java.awt.Dimension(APPLET_WIDTH, APPLET_HEIGHT));
    frame.add(applet, java.awt.BorderLayout.CENTER);
    frame.setResizable(false);
    frame.pack();
    frame.setLocationRelativeTo(null);
    frame.setVisible(true);
    Thread.sleep(250);
    applet.start();
  }
}
